{===========================================================================
ZOOMLER v 1.1

Zoomer & Scroller -- two in one.
Mouse interface only. 
This control is handy to use with any zoomable objects -- plots, bitmaps etc.

Freeware.

In the case of any changes of the source code please send me the copy.

"As is". No responsibility in any case.

(c) 1997 Alexander Rublinetsky (rublin@iname.com)

Dec 22, 1997

============================================================================}
unit Zoomler;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

const
  sc_LEFTPAGE = 0;
  sc_RIGHTPAGE = 1;
  sc_LEFTSIZE = 2;
  sc_RIGHTSIZE = 3;
  sc_MOVE = 4;
  sc_LEFTMOVE = 5;
  sc_RIGHTMOVE = 6;
  sc_NONE = 7;

  TimerDelay = 500;
  TimerRate  = 50;

type
  TOrientation = (Horizontal, Vertical);

  TZoomler = class(TGraphicControl)
  private
    FSL,FSR      :Integer;
    FMin,FMax,
    FStart,
    FFinish,
    FStep        : Double;
    Status       : (st_NONE, st_LEFT, st_RIGHT, st_BOTH);
    XS           : integer;
    FDblClk      : boolean;
    FScrolling   : TNotifyEvent;
    FZooming     : TNotifyEvent;
    FChanged     : TNotifyEvent;
    FOrientation : TOrientation;
    Timer        : TTimer;
    hButton      : TMouseButton;
    hShift       : TShiftState;
    hX, hY       : Integer;
    LeftMoving,
    RightMoving  : Boolean;
    procedure ProcessMove(X:integer);
    function WherePoint(X,Y:integer):byte;
    procedure CalcScreen;
    procedure DoLeftMove;
    procedure DoRightMove;
    procedure DoLeftPage(X:integer);
    procedure DoRightPage(X:integer);
    procedure SetMin(m:double);
    procedure SetMax(m:double);
    procedure SetStart(m:double);
    procedure SetFinish(m:double);
    Procedure SetOrientation(aOrientation : TOrientation);
    Procedure OnTimer(Sender : TOBject);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
                        X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState;
                        X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
                        X, Y: Integer); override;
    procedure DblClick; override;
    procedure Paint; override;
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy;Override;
  published
    property Min:Double read FMin write SetMin;
    property Max:Double read FMax write SetMax;
    property Start:Double read FStart write SetStart;
    property Finish:Double read FFinish write SetFinish;
    property DblClkEnable:boolean read FDblClk write FDblClk default True;
    property Align;
    property Visible;
    property OnScrolling:TNotifyEvent read FScrolling write FScrolling;
    property OnZooming:TNotifyEvent read FZooming write FZooming;
    property OnChanged:TNotifyEvent read FChanged write FChanged;
    property Orientation : TOrientation Read FOrientation Write SetOrientation;
  end;

function DMin(a,b:double):double;
function DMax(a,b:double):double;

procedure UpRect(C:TCanvas; x1,y1,x2,y2:integer);
procedure DownRect(C:TCanvas; x1,y1,x2,y2:integer);

procedure Register;

implementation

procedure UpRect(C:TCanvas; x1,y1,x2,y2:integer);
begin
   with C do begin
     Brush.Color:=clSilver;
     FillRect(rect(x1,y1,x2,y2));
     Pen.Color:=clWhite;
     MoveTo(x1,y2);LineTo(x1,y1);LineTo(x2,y1);
     Pen.Color:=cl3DDkShadow;
     MoveTo(x2,y1+1);LineTo(x2,y2);LineTo(x1-1,y2);
   end;
end;

procedure DownRect(C:TCanvas; x1,y1,x2,y2:integer);
begin
   with C do begin
     Brush.Color:=clSilver;
     FillRect(rect(x1,y1,x2,y2));
     Pen.Color:=clWhite;
     MoveTo(x2,y1);LineTo(x2,y2);LineTo(x1,y2);
     Pen.Color:=cl3DDkShadow;
     MoveTo(x1,y2);LineTo(x1,y1);LineTo(x2,y1);
   end;
end;

function DMin(a,b:double):double;
begin
    if a<b then Result:=a else Result:=b;
end;

function DMax(a,b:double):double;
begin
    if a>b then Result:=a else Result:=b;
end;

constructor TZoomler.Create(Owner: TComponent);
begin
   inherited Create(Owner);
   ControlStyle := ControlStyle + [csOpaque];
   FMin:=1.0; FMax:=100.0;
   FStart:=30.0; FFinish:=70.0;
   FStep:=1.0;
   Width:=200; Height:=14;
   CalcScreen;
   FDblClk:=True;
   Status:=st_NONE;
   Orientation := Horizontal;
   Timer := TTimer.Create(Self);
   Timer.OnTimer := OnTimer;
   Timer.Enabled := False;
end;

Destructor TZoomler.Destroy;
Begin
   Timer.Destroy;
   Inherited;
End;

procedure TZoomler.SetMin(m:double);
begin
   if m>=FMax then FMin:=FMax-FStep else FMin:=m;
   CalcScreen;
   Repaint;
end;

procedure TZoomler.SetMax(m:double);
begin
   if m<=FMin then FMax:=FMin+FStep else FMax:=m;
   CalcScreen;
   Repaint;
end;

procedure TZoomler.SetStart(m:double);
begin
   if m>=FFinish then FStart:=FFinish-FStep;
   if m<FMin then FStart:=FMin else FStart:=m;
   CalcScreen;
   Repaint;
end;

procedure TZoomler.SetFinish(m:double);
begin
   if m<=FStart then FFinish:=FStart+FStep;
   if m>FMax then FFinish:=FMax else FFinish:=m;
   CalcScreen;
   Repaint;
end;

Procedure TZoomler.SetOrientation;
Begin
   FOrientation := aOrientation;
   Repaint;
End;

Procedure TZoomler.OnTimer;
Begin
   Timer.Interval := TimerRate;
   MouseDown(hButton, hShift, hx, hy);
End;

procedure TZoomler.DblClick;
begin
   inherited;
   if not FDblClk then Exit;
   FStart:=FMin;
   FFinish:=FMax;
   CalcScreen;
   Repaint;
   if Assigned(FChanged) then FChanged(Self);
end;

procedure TZoomler.Paint;
var Buffer : TBitmap;
begin
   Buffer:=TBitmap.Create;
   Buffer.Width:=Width;
   Buffer.Height:=Height;
   CalcScreen;
   with Buffer.Canvas do begin
      DownRect(Buffer.Canvas,0,0,Width-1,Height-1);
      Brush.Color:=$E0E0E0;
      Pen.Color:=$E0E0E0;
      Rectangle(2,2,Width-2,Height-2);
      Case Orientation Of
         Horizontal : Begin
                         UpRect(Buffer.Canvas,FSL,2,FSR,Height-3);
                         DownRect(Buffer.Canvas,FSL+10,3,FSR-10,Height-4);
                         If LeftMoving Then
                            DownRect(Buffer.Canvas,1,2,11,Height-3)
                         Else
                            UpRect(Buffer.Canvas,1,2,11,Height-3);
                         If RightMoving Then
                           DownRect(Buffer.Canvas,Width-11,2,Width-2,Height-3)
                         Else
                           UpRect(Buffer.Canvas,Width-11,2,Width-2,Height-3);
                         Pen.Color:=clGray;
                         Brush.Color:=clGray;
                         MoveTo(FSL+5,3); LineTo(FSL+5,Height-4);
                         MoveTo(FSR-5,3); LineTo(FSR-5,Height-4);
                         Polygon([
                            Point(9,3),
                            Point(9,Height-4),
                            Point(2,(Height-4) div 2+2)]);
                         Polygon([
                            Point(Width-9,3),
                            Point(Width-9,Height-4),
                            Point(Width-3,(Height-4) div 2+2)]);
                      End;
         Vertical   : Begin
                         UpRect(Buffer.Canvas,2,FSL,Width - 3,FSR);
                         DownRect(Buffer.Canvas,3,FSL+10,Width - 4,FSR-10);
                         If LeftMoving Then
                            DownRect(Buffer.Canvas,2,1,width - 3,11)
                         Else
                            UpRect(Buffer.Canvas,2,1,width - 3,11);
                         If RightMoving Then
                            DownRect(Buffer.Canvas,2,Height-11,Width-3,Height-2)
                         Else
                            UpRect(Buffer.Canvas,2,Height-11,Width-3,Height-2);
                         Pen.Color:=clGray;
                         Brush.Color:=clGray;
                         MoveTo(3,FSL+5); LineTo(Width-4,FSL+5);
                         MoveTo(3,FSR-5); LineTo(Width-4,FSR-5);
                         Polygon([
                            Point(3,9),
                            Point(Width-4,9),
                            Point((width-4) div 2+2,2)]);
                         Polygon([
                            Point(3,Height-9),
                            Point(Width-4,Height-9),
                            Point((Width-4) div 2+2,Height-3)]);
                      End;
      End;
   end;
   Canvas.CopyMode:=cmSrcCopy;
   Canvas.Draw(0,0,Buffer);
   Buffer.Destroy;
end;

procedure TZoomler.CalcScreen; {double to screen}
var coef:double;
    osl,osr:integer;
begin
   Case Orientation Of
      Horizontal : coef:=(width-24)/(FMax-FMin);
      Vertical   : coef:=(Height-24)/(FMax-FMin);
   End;
   osl:=FSL;
   osr:=FSR;
   FSL:=12+Round((FStart-FMin)*coef);
   Case Orientation Of
      Horizontal : FSR:=Width-12-Round((FMax-FFinish)*coef);
      Vertical   : FSR:=Height-12-Round((FMax-FFinish)*coef);
   End;
   if (FSR-FSL)<30 then begin
      if abs(FSR-osr)>abs(FSL-osl) then FSR:=FSL+30 else
      if abs(FSL-osl)>abs(FSR-osr) then FSL:=FSR-30 else begin
        FSL:=osl;
        FSR:=osr;
      end;
   end;
end;

procedure TZoomler.DoLeftMove; {one step to left}
var d:double;
    i:integer;
begin
    if FStart>=FMin+FStep then d:=FStep
                          else d:=FStart-FMin;
    FStart:=FStart-d;
    FFinish:=FFinish-d;
    CalcScreen;
    Repaint;
    if Assigned(FChanged) then FChanged(Self);
end;

procedure TZoomler.DoRightMove; {one step to right}
var d:double;
    i:integer;
begin
    if FFinish<=FMax-FStep then d:=FStep
                           else d:=FMax-FFinish;
    FStart:=FStart+d;
    FFinish:=FFinish+d;
    CalcScreen;
    Repaint;
    if Assigned(FChanged) then FChanged(Self);
end;

procedure TZoomler.DoLeftPage(X:integer); {proportionally}
var d:double;
begin
    d:=FStart;
    FStart:=FStart-(FStart-FMin)/(FSL-11)*(FSL-x);
    FFinish:=FFinish-(d-FStart);
    CalcScreen;
    Repaint;
    if Assigned(FChanged) then FChanged(Self);
end;

procedure TZoomler.DoRightPage(X:integer); {proportionally}
var d:double;
begin
    d:=FFinish;
    Case Orientation Of
       Horizontal : FFinish:=FFinish+(FMax-FFinish)/(Width-11-FSR)*(x-FSR);
       Vertical   : FFinish:=FFinish+(FMax-FFinish)/(Height-11-FSR)*(x-FSR);
    End;
    FStart:=FStart+(Finish-d);
    CalcScreen;
    Repaint;
    if Assigned(FChanged) then FChanged(Self);
end;

procedure TZoomler.MouseDown(Button: TMouseButton; Shift: TShiftState;
                              X, Y: Integer);
var p:byte;
    m : Integer;
begin
   inherited;
   Case Orientation of
      Horizontal : m:=X;
      Vertical   : m:=Y;
   End;
   XS:=m;
   p:=WherePoint(X,Y);
   case p of
     sc_LEFTSIZE  : Status:=st_LEFT;
     sc_RIGHTSIZE : Status:=st_RIGHT;
     sc_LEFTMOVE  : Begin DoLeftMove;LeftMoving := True;End;
     sc_RIGHTMOVE : Begin DoRightMove;RightMoving := True;End;
     sc_MOVE      : Status:=st_BOTH;
     sc_LEFTPAGE  : DoLeftPage(m);
     sc_RIGHTPAGE : DoRightPage(m);
     else           Status:=st_NONE;
   end;
   If p in [sc_LEFTMOVE, sc_RIGHTMOVE] Then Begin
      if Not(Timer.Enabled) Then Timer.Interval := TimerDelay;
      hx := x;
      hy := y;
      hButton := Button;
      hShift := Shift;
      Timer.Enabled := True;
   End;
end;

procedure TZoomler.MouseUp(Button: TMouseButton; Shift: TShiftState;
                              X, Y: Integer);
begin
   inherited;
   Timer.Enabled := False;
   LeftMoving := False;
   RightMoving := False;
   if Status<>st_NONE then
      if Assigned(FChanged) then FChanged(Self);
   Repaint;
   Status:=st_NONE;
end;

procedure TZoomler.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
   inherited;
   case WherePoint(X,Y) of
      sc_LEFTSIZE,sc_RIGHTSIZE : Case Orientation Of
                                    Horizontal : Cursor:=crSizeWE;
                                    Vertical   : Cursor:=crSizeNS;
                                 End;
      sc_MOVE : Cursor:=crHandPoint;
      else Cursor:=crDefault;
   end;
   if Status<>st_NONE then Begin
      Case orientation Of
         Horizontal : ProcessMove(X);
         Vertical   : ProcessMove(Y);
      End;
   end;
end;

procedure TZoomler.ProcessMove(X:integer);
var d:double;
    coef:double;
begin
   if X=XS then Exit;
   Case Orientation Of
      Horizontal : coef:=(Width-24)/(FMax-FMin);
      Vertical   : coef:=(Height-24)/(FMax-FMin);
   End;
   d:=(X-XS)/coef;
   case Status of
      st_BOTH  : begin
                   if d<0 then d:=DMax(d,FMin-FStart)
                          else d:=DMin(d,FMax-FFinish);
                   FStart:=FStart+d;
                   FFinish:=FFinish+d;
                   if Assigned(FScrolling) then FScrolling(Self);
                 end;
      st_LEFT  : begin
                   if d<0 then FStart:=FStart+DMax(d,FMin-FStart)
                          else FStart:=FStart+DMin(d,FFinish-FStart);
                   if Assigned(FZooming) then FZooming(Self);
                 end;
      st_RIGHT  : begin
                   if d>0 then FFinish:=FFinish+DMin(d,FMax-FFinish)
                          else FFinish:=FFinish+DMax(d,FStart-FFinish);
                   if Assigned(FZooming) then FZooming(Self);
                  end;
   end;
   XS:=X;
   Repaint;
end;

function TZoomler.WherePoint(X,Y:integer):byte;
begin
   Case Orientation of
      Horizontal : Begin
                      if PtInRect(rect(1,3,11,Height-3),Point(X,Y)) then
                         Result:=sc_LEFTMOVE
                      else
                      if PtInRect(rect(12,3,FSL-1,Height-3),Point(X,Y)) then
                         Result:=sc_LEFTPAGE
                      else
                      if PtInRect(rect(FSL+1,3,FSL+11,Height-3),Point(X,Y)) then
                         Result:=sc_LEFTSIZE
                      else
                      if PtInRect(rect(FSL+12,3,FSR-12,Height-3),Point(X,Y)) then
                         Result:=sc_MOVE
                      else
                      if PtInRect(rect(FSR-11,3,FSR-1,Height-3),Point(X,Y)) then
                         Result:=sc_RIGHTSIZE
                      else
                      if PtInRect(rect(FSR+1,3,Width-12,Height-3),Point(X,Y)) then
                         Result:=sc_RIGHTPAGE
                      else
                      if PtInRect(rect(Width-11,3,Width-1,Height-3),Point(X,Y)) then
                         Result:=sc_RIGHTMOVE
                      else
                         Result:=sc_None;
                    End;
      Vertical   : Begin
                      if PtInRect(rect(3,1,width-3,11),Point(X,Y)) then
                         Result:=sc_LEFTMOVE
                      else
                      if PtInRect(rect(3,12,Width-3,FSL-1),Point(X,Y)) then
                         Result:=sc_LEFTPAGE
                      else
                      if PtInRect(rect(3,FSL+1,Width-3,FSL+11),Point(X,Y)) then
                         Result:=sc_LEFTSIZE
                      else
                      if PtInRect(rect(3,FSL+12,Width-3,FSR-12),Point(X,Y)) then
                         Result:=sc_MOVE
                      else
                      if PtInRect(rect(3,FSR-11,Width-3,FSR-1),Point(X,Y)) then
                         Result:=sc_RIGHTSIZE
                      else
                      if PtInRect(rect(3,FSR+1,Width-3,Height-12),Point(X,Y)) then
                         Result:=sc_RIGHTPAGE
                      else
                      if PtInRect(rect(3,Height-11,Width-3,Height-1),Point(X,Y)) then
                         Result:=sc_RIGHTMOVE
                      else
                         Result:=sc_None;
                    End;
      End;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TZoomler]);
end;

end.
